/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2012 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Sam Lantinga
    slouken@libsdl.org
*/
#include "SDL_config.h"

#include <sys/keycodes.h>
#include <sys/termmode.h>

#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <dispd.h>
#if __has_include(<display.h>)
#define DISPLAY
#include <display.h>
#endif
#include <fcntl.h>
#include <unistd.h>

#include "SDL_video.h"
#include "../SDL_sysvideo.h"
#include "../../events/SDL_sysevents.h"
#include "../../events/SDL_events_c.h"

#include "SDL_dispd.h"

static int DispD_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **DispD_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
static SDL_Surface *DispD_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
static int DispD_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
static void DispD_VideoQuit(_THIS);
static int DispD_AllocHWSurface(_THIS, SDL_Surface *surface);
static int DispD_LockHWSurface(_THIS, SDL_Surface *surface);
static void DispD_UnlockHWSurface(_THIS, SDL_Surface *surface);
static void DispD_FreeHWSurface(_THIS, SDL_Surface *surface);
static void DispD_SetCaption(_THIS, const char *title, const char *icon);
static void DispD_UpdateRects(_THIS, int numrects, SDL_Rect *rects);
static void DispD_PumpEvents(_THIS);
static void DispD_InitOSKeymap(_THIS);

static int DispD_Available(void)
{
	return 1;
}

static void DispD_DeleteDevice(SDL_VideoDevice *device)
{
#ifdef DISPLAY
	if ( device->hidden->connection ) {
		display_destroy_window(device->hidden->connection,
		                       device->hidden->window_id);
		display_disconnect(device->hidden->connection);
		device->hidden->connection = NULL;
	}
#endif
	if ( device->hidden->fbinfo ) {
		dispd_finish_render(device->hidden->fbinfo);
		device->hidden->fbinfo = NULL;
	}
	if ( device->hidden->window ) {
		dispd_destroy_window(device->hidden->window);
		device->hidden->window = NULL;
	}
	if ( device->hidden->session ) {
		dispd_detach_session(device->hidden->session);
		device->hidden->session = NULL;
	}
	if ( 0 < device->hidden->tty_fd ) {
		close(device->hidden->tty_fd);
		device->hidden->tty_fd = -1;
	}
	SDL_free(device->hidden);
	SDL_free(device);
}

static SDL_VideoDevice *DispD_CreateDevice(int devindex)
{
	SDL_VideoDevice *device;

	/* Initialize all variables that we clean on shutdown */
	device = (SDL_VideoDevice *)SDL_malloc(sizeof(SDL_VideoDevice));
	if ( device ) {
		SDL_memset(device, 0, (sizeof *device));
		device->hidden = (struct SDL_PrivateVideoData *)
				SDL_malloc((sizeof *device->hidden));
	}
	if ( (device == NULL) || (device->hidden == NULL) ) {
		SDL_OutOfMemory();
		if ( device ) {
			SDL_free(device);
		}
		return(0);
	}
	SDL_memset(device->hidden, 0, (sizeof *device->hidden));

#ifdef DISPLAY
	if ( (device->hidden->connection = display_connect_default()) ) {
		device->hidden->disconnected = 0;
		device->hidden->window_id = 0;
		device->hidden->window_width = 0;
		device->hidden->window_height = 0;
		display_create_window(device->hidden->connection,
		                      device->hidden->window_id);
	} else {
#endif
		static int has_initialized_dispd = 0;
		if ( !has_initialized_dispd ) {
			if ( !dispd_initialize(NULL, NULL) ) {
				return(0);
			}
			has_initialized_dispd = 1;
		}
		if ( (device->hidden->tty_fd = open("/dev/tty", O_RDONLY)) < 0 ) {
			DispD_DeleteDevice(device);
			return(0);
		}

		if ( (device->hidden->session = dispd_attach_default_session()) == NULL ) {
			DispD_DeleteDevice(device);
			return(0);
		}

		if ( !(dispd_session_setup_game_rgba(device->hidden->session)) ) {
			DispD_DeleteDevice(device);
			return(0);
		}

		if ( (device->hidden->window =
			  dispd_create_window_game_rgba(device->hidden->session)) == NULL ) {
			DispD_DeleteDevice(device);
			return(0);
		}

		if ( (device->hidden->fbinfo =
			   dispd_begin_render(device->hidden->window)) == NULL ) {
			DispD_DeleteDevice(device);
			return(0);
		}
#ifdef DISPLAY
	}
#endif

	device->hidden->current_mode.x = 0;
	device->hidden->current_mode.y = 0;
#ifdef DISPLAY
	if ( device->hidden->connection ) {
		device->hidden->current_mode.w = 800;
		device->hidden->current_mode.h = 600;
	} else {
#endif
		device->hidden->current_mode.w =
			dispd_get_framebuffer_width(device->hidden->fbinfo);
		device->hidden->current_mode.h =
			dispd_get_framebuffer_height(device->hidden->fbinfo);
#ifdef DISPLAY
	}
#endif
	device->hidden->mode_list[0] = &device->hidden->current_mode;
	device->hidden->mode_list[1] = NULL;

	/* Set the function pointers */
	device->VideoInit = DispD_VideoInit;
	device->ListModes = DispD_ListModes;
	device->SetVideoMode = DispD_SetVideoMode;
	device->CreateYUVOverlay = NULL;
	device->SetColors = DispD_SetColors;
	device->UpdateRects = DispD_UpdateRects;
	device->VideoQuit = DispD_VideoQuit;
	device->AllocHWSurface = DispD_AllocHWSurface;
	device->CheckHWBlit = NULL;
	device->FillHWRect = NULL;
	device->SetHWColorKey = NULL;
	device->SetHWAlpha = NULL;
	device->LockHWSurface = DispD_LockHWSurface;
	device->UnlockHWSurface = DispD_UnlockHWSurface;
	device->FlipHWSurface = NULL;
	device->FreeHWSurface = DispD_FreeHWSurface;
	device->SetCaption = DispD_SetCaption;
	device->SetIcon = NULL;
	device->IconifyWindow = NULL;
	device->GrabInput = NULL;
	device->GetWMInfo = NULL;
	device->InitOSKeymap = DispD_InitOSKeymap;
	device->PumpEvents = DispD_PumpEvents;

	device->free = DispD_DeleteDevice;

	return device;
}

int DispD_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
	vformat->BitsPerPixel = 32;
	vformat->BytesPerPixel = vformat->BitsPerPixel/8;
	return(0);
}

SDL_Rect **DispD_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
	// TODO: Return NULL if the format isn't 32-bit supported.
#ifdef DISPLAY
	if ( this->hidden->connection ) {
		// TODO: qemu seems to pick too little a resolution due to this.
		return((SDL_Rect **)-1);
	}
#endif
	return(this->hidden->mode_list);
}

SDL_Surface *DispD_SetVideoMode(_THIS, SDL_Surface *current,
				int width, int height, int bpp, Uint32 flags)
{
	uint8_t *data;
	size_t pitch;

	bpp = 32;

#ifdef DISPLAY
	if ( this->hidden->connection ) {
		current->flags = SDL_RESIZABLE;
		size_t size = (size_t)width * (size_t)width * (bpp / 8);
		data = SDL_malloc(size);
		if ( !data )
			return(NULL);
		this->hidden->window_width = width;
		this->hidden->window_height = height;
		display_resize_window(this->hidden->connection,
		                      this->hidden->window_id, width, height);
		pitch = (size_t) width * (bpp / 8);
	} else {
#endif
		data = dispd_get_framebuffer_data(this->hidden->fbinfo);
		width = dispd_get_framebuffer_width(this->hidden->fbinfo);
		height = dispd_get_framebuffer_height(this->hidden->fbinfo);
		pitch = dispd_get_framebuffer_pitch(this->hidden->fbinfo);
		current->flags = SDL_FULLSCREEN;
#ifdef DISPLAY
	}
#endif

	int y;
	for ( y = 0; y < height; y++ )
	{
		uint8_t *line = data + y * pitch;
		size_t line_size = width * bpp / 8;
		SDL_memset(line, 0, line_size);
	}

	// TODO: What's with these assertions? Should this not be an error? How is
	//       this ensured as a precondition?
	assert(current->format);
	assert(current->format->BitsPerPixel == 32);
	current->pitch = pitch;
	// TODO: Memory leak of old buffer?
	current->pixels = data;
	current->w = width;
	current->h = height;

	return(current);
}

/* We don't actually allow hardware surfaces other than the main one */
static int DispD_AllocHWSurface(_THIS, SDL_Surface *surface)
{
	return(-1);
}

static void DispD_FreeHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

static void DispD_SetCaption(_THIS, const char *title, const char *icon)
{
#ifdef DISPLAY
	if ( this->hidden->connection) {
		display_title_window(this->hidden->connection,
		                     this->hidden->window_id, title);
	}
#endif
}

/* We need to wait for vertical retrace on page flipped displays */
static int DispD_LockHWSurface(_THIS, SDL_Surface *surface)
{
	return(0);
}

static void DispD_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

static void DispD_UpdateRects(_THIS, int numrects, SDL_Rect *rects)
{
#ifdef DISPLAY
	if ( this->hidden->connection) {
		for ( size_t i = 3; i < (size_t)SDL_VideoSurface->w * (size_t)SDL_VideoSurface->h * 4; i += 4 )
			((unsigned char*)SDL_VideoSurface->pixels)[i] = 255;
		display_render_window(this->hidden->connection, this->hidden->window_id,
		                      0, 0, SDL_VideoSurface->w, SDL_VideoSurface->h,
		                      SDL_VideoSurface->pixels);
		display_show_window(this->hidden->connection, this->hidden->window_id);
		return;
	}
#endif
	uint8_t* old_data = dispd_get_framebuffer_data(this->hidden->fbinfo);
	if ( !dispd_finish_render(this->hidden->fbinfo) ) {
		abort();
	}
	if ( !(this->hidden->fbinfo = dispd_begin_render(this->hidden->window)) ) {
		abort();
	}
	uint8_t* new_data = dispd_get_framebuffer_data(this->hidden->fbinfo);
	assert(old_data == new_data);
}

int DispD_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
	/* do nothing of note. */
	return(1);
}

/* Note:  If we are terminated, this could be called in the middle of
   another SDL video routine -- notably UpdateRects.
*/
void DispD_VideoQuit(_THIS)
{
	this->screen->pixels = NULL;
}

static int TranslateKey(int abskbkey)
{
	switch ( abskbkey )
	{
	case KBKEY_ESC: return SDLK_ESCAPE;
	case KBKEY_ENTER: return SDLK_RETURN;
	case KBKEY_BKSPC: return SDLK_BACKSPACE;
	case KBKEY_TAB: return SDLK_TAB;
	case KBKEY_SPACE: return SDLK_SPACE;
	case KBKEY_LEFT: return SDLK_LEFT;
	case KBKEY_RIGHT: return SDLK_RIGHT;
	case KBKEY_UP: return SDLK_UP;
	case KBKEY_DOWN: return SDLK_DOWN;
	case KBKEY_LCTRL: return SDLK_LCTRL;
	case KBKEY_RCTRL: return SDLK_RCTRL;
	case KBKEY_LSHIFT: return SDLK_LSHIFT;
	case KBKEY_RSHIFT: return SDLK_RSHIFT;
	case KBKEY_F1: return SDLK_F1;
	case KBKEY_F2: return SDLK_F2;
	case KBKEY_F3: return SDLK_F3;
	case KBKEY_F4: return SDLK_F4;
	case KBKEY_F5: return SDLK_F5;
	case KBKEY_F6: return SDLK_F6;
	case KBKEY_F7: return SDLK_F7;
	case KBKEY_F8: return SDLK_F8;
	case KBKEY_F9: return SDLK_F9;
	case KBKEY_F10: return SDLK_F10;
	case KBKEY_INSERT: return SDLK_INSERT;
	case KBKEY_DELETE: return SDLK_DELETE;
	case KBKEY_PGUP: return SDLK_PAGEUP;
	case KBKEY_PGDOWN: return SDLK_PAGEDOWN;
	case KBKEY_HOME: return SDLK_HOME;
	case KBKEY_END: return SDLK_END;
	// TODO: SDLK_PAUSE
	case KBKEY_NUM1: return SDLK_1;
	case KBKEY_NUM2: return SDLK_2;
	case KBKEY_NUM3: return SDLK_3;
	case KBKEY_NUM4: return SDLK_4;
	case KBKEY_NUM5: return SDLK_5;
	case KBKEY_NUM6: return SDLK_6;
	case KBKEY_NUM7: return SDLK_7;
	case KBKEY_NUM8: return SDLK_8;
	case KBKEY_NUM9: return SDLK_9;
	case KBKEY_NUM0: return SDLK_0;
	case KBKEY_A: return SDLK_a;
	case KBKEY_B: return SDLK_b;
	case KBKEY_C: return SDLK_c;
	case KBKEY_D: return SDLK_d;
	case KBKEY_E: return SDLK_e;
	case KBKEY_F: return SDLK_f;
	case KBKEY_G: return SDLK_g;
	case KBKEY_H: return SDLK_h;
	case KBKEY_I: return SDLK_i;
	case KBKEY_J: return SDLK_j;
	case KBKEY_K: return SDLK_k;
	case KBKEY_L: return SDLK_l;
	case KBKEY_M: return SDLK_m;
	case KBKEY_N: return SDLK_n;
	case KBKEY_O: return SDLK_o;
	case KBKEY_P: return SDLK_p;
	case KBKEY_Q: return SDLK_q;
	case KBKEY_R: return SDLK_r;
	case KBKEY_S: return SDLK_s;
	case KBKEY_T: return SDLK_t;
	case KBKEY_U: return SDLK_u;
	case KBKEY_V: return SDLK_v;
	case KBKEY_W: return SDLK_w;
	case KBKEY_X: return SDLK_x;
	case KBKEY_Y: return SDLK_y;
	case KBKEY_Z: return SDLK_z;
	case KBKEY_SYM11: return SDLK_SLASH; // TODO: Information loss.
	// TODO: More stuff here!
	}
	return 0;
}

#ifdef DISPLAY
static void on_disconnect(void* ctx)
{
	struct SDL_PrivateVideoData* hidden = (struct SDL_PrivateVideoData*) ctx;
	hidden->disconnected = 1;
}

static void on_quit(void* ctx, uint32_t window_id)
{
	struct SDL_PrivateVideoData* hidden = (struct SDL_PrivateVideoData*) ctx;
	hidden->disconnected = 1;
}

static void on_resize(void* ctx, uint32_t window_id, uint32_t width,
                      uint32_t height)
{
	// Discard notification from when the window is created with zero size.
	if ( width == 0 || height == 0 )
		return;
	struct SDL_PrivateVideoData* hidden = (struct SDL_PrivateVideoData*) ctx;
	if ( hidden->window_width != width || hidden->window_height != height ) {
		hidden->window_width = width;
		hidden->window_height = height;
		SDL_PrivateResize(width, height);
	}
}

static void on_keyboard(void* ctx, uint32_t window_id, uint32_t codepoint)
{
	int kbkey = KBKEY_DECODE(codepoint);
	int abskbkey = kbkey < 0 ? -kbkey : kbkey;
	int key = TranslateKey(abskbkey);
	SDL_keysym keysym;
	keysym.scancode = abskbkey;
	keysym.sym = key;
	keysym.mod = 0;
	keysym.unicode = 0;
	SDL_PrivateKeyboard(kbkey < 0 ? SDL_RELEASED : SDL_PRESSED, &keysym);
}
#endif

void DispD_PumpEvents(_THIS)
{
#ifdef DISPLAY
	if ( this->hidden->connection) {
		struct display_event_handlers handlers;
		memset(&handlers, 0, sizeof(handlers));
		handlers.context = this->hidden;
		handlers.disconnect_handler = on_disconnect;
		handlers.quit_handler = on_quit;
		handlers.resize_handler = on_resize;
		handlers.keyboard_handler = on_keyboard;
		while ( !this->hidden->disconnected ) {
			if ( display_poll_event(this->hidden->connection, &handlers) < 0 )
				break;
		}
		return;
	}
#endif

	// Read the keyboard input from the user.
	const unsigned termmode = TERMMODE_KBKEY
	                        | TERMMODE_UNICODE
	                        | TERMMODE_SIGNAL
	                        | TERMMODE_NONBLOCK;
	if ( settermmode(0, termmode) ) {
		return;
	}
	uint32_t codepoint;
	ssize_t numbytes;
	while ( 0 < (numbytes = read(0, &codepoint, sizeof(codepoint))) )
	{
		int kbkey = KBKEY_DECODE(codepoint);
		int abskbkey = kbkey < 0 ? -kbkey : kbkey;
		int key = TranslateKey(abskbkey);
		SDL_keysym keysym;
		keysym.scancode = abskbkey;
		keysym.sym = key;
		keysym.mod = 0;
		keysym.unicode = 0;
		SDL_PrivateKeyboard(kbkey < 0 ? SDL_RELEASED : SDL_PRESSED, &keysym);
	}
}

void DispD_InitOSKeymap(_THIS)
{
	/* do nothing. */
}

VideoBootStrap DispD_bootstrap = {
  "dispd", "DispD",
  DispD_Available, DispD_CreateDevice
};
